# Copyright UCA/CNRS/Inria
# Contributor(s): Cedric Dubois 2022
#
# cedric.dubois@inria.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#
#
import numpy as np
from scipy.spatial import distance_matrix
import itertools
import warnings
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.class_weight import compute_sample_weight
from sklearn.neighbors import KNeighborsClassifier as KNNC
from scipy.sparse import coo_array, csr_matrix, csr_array
# from _utils import _binary_search_perplexity, _binary_search_perplexity_umap_like

EPSILON = 1e-8
NPY_INFINITY = np.inf
PERPLEXITY_TOLERANCE = 1e-5

def compute_rank(D):
    n = D.shape[0] - 1
    B = np.zeros([n, n])
    for i, j in itertools.product(range(n), range(n)):
        B[i, j] = (D[i, -1] ** 2. + D[j, -1] ** 2. - D[i, j] ** 2.) / 2.
    #
    return np.linalg.matrix_rank(B, hermitian=True)  # TODO: Check if always hermitian

class T2T(BaseEstimator, TransformerMixin):
    """
    Transformation 2 Targets
    """
    def __init__(self,
                 gamma=1,
                 k_nn=100,
                 class_weight=None,
                 raise_warning=True,
                 threshold=1e-9,
                 verbose=0,
                 ):
        """
        Parameters
        ----------
        :param scaling_factor: float, kernel width, sigma
        :param k_nn: int, number of nearest neighbors to use for estimating the translations
        :param class_weight: str, None or 'balanced'
        :param raise_warning: bool
        :param threshold: float
        :param verbose: int
        """
        #
        # self.scaling_factor = scaling_factor
        self.gamma = gamma
        self.k_nn = k_nn
        self.raise_warning = raise_warning
        self.threshold = threshold
        self.class_weight = class_weight
        self.verbose = verbose

        self.k_nn_model = None
        self.saved_X = False
        self.saved_knn = 0

    def fit(self, X, y, targets='canonical-basis'):
        """
        Fit X transformation based on y.
        :param X: array, shape (n_samples, n_features). Samples to fit
        :param y: array, shape (n_samples). Target labels
        :param sample_target: array (n_samples), specific targets
        :return:
        """
        self.classes_ = np.unique(y)
        self.n_classes_ = self.classes_.shape[0]
        if isinstance(targets, str):
            if targets=='barycenter':
                print("Using class barycenter as targets")
                targets = np.empty([self.n_classes_, X.shape[1]])
                k=0
                for lab in self.classes_:
                    targets[k] = np.mean(X[y==lab], axis=0)
                    k+=1
            elif targets=='equidistant' or targets=='eq' or targets=='simplex':
                dimension = self.n_classes_ - 1
                if self.verbose>0:
                    print(f"Using {dimension}-regular-simplex as targets")
                # From package rn_simplex :
                targets = np.empty((dimension + 1, dimension), dtype=np.float64)
                targets[:dimension, :] = np.eye(dimension)
                targets[dimension, :] = (1.0 + np.sqrt(dimension + 1.0)) / dimension
            elif targets=='canonical-basis':
                if self.verbose > 0:
                    print(f"Using n={self.n_classes_} canonical basis as targets")
                targets = np.eye(self.n_classes_)

        elif isinstance(targets, csr_matrix | csr_array | np.ndarray):
            print("Using given targets")
        else:
            raise NotImplementedError("Provide targets as scipy csr_matrix or scipy csr_array or numpy ndarray or string")
        #
        self.sample_classes_train = y
        #
        self.samples_train = X
        self.n_samples = X.shape[0]
        #
        self.sample_targets = targets[y]
        self.targets = targets
        #
        k_nn = self.k_nn
        k_max = self.n_samples
        if k_nn == None or k_nn > k_max:
            k_nn = k_max

        # TODO: Change this classifer to a simple graph construction (but this method seems faster)
        self.k_nn_model = KNNC(n_neighbors=k_nn,
                               n_jobs=-1,
                               leaf_size=40,
                               algorithm='auto', )
        self.k_nn_model.fit(self.samples_train, self.sample_classes_train)

        return self

    def fit_local_gamma_y(self, X, y, targets='canonical-basis', k_nn_local_gamma=10):
        """
        Fit + local gamma per learning sample estimates
        :param X:
        :param y:
        :return:
        """
        self.fit(X, y, targets=targets)
        if self.verbose:
            print("--- Searching for neighbors with sklearn for local gamma estimation (on y)")
        distances_y, _ = self.k_nn_model.kneighbors(X, n_neighbors=k_nn_local_gamma)
        self.r_y = distances_y[:, -1]

    def transform(self, X, k_nn=None, gamma=None, loocv=False):
        """
        transform samples according to training data.
        Parameters
        ----------
        :param X: samples to transform
        :param k_nn (optional): int, number of nearest neighbors to use for estimating the translations
        :param gamma (optional): float, kernel inverse scaling factor
        :return: transformed samples
        """
        if self.k_nn==None:
            return self.transform_exact(X)
        else:
            return self.transform_approx(X, k_nn, gamma=gamma,  loocv=loocv)

    def transform_exact(self, X):
        """ Weighted knn like
        :param X:
        :return:
        """
        warnings.warn("Should not be used for now.")
        if self.raise_warning:
            warnings.warn("k_nn param not provided, computing transformation from all neighbors. "
                          f"This compute the Gram matrix {self.samples_train.shape[0]} x {X.shape[0]}")

        assert X.shape[1] == self.samples_train.shape[1]
        #
        if self.verbose:
            print("Computing Query radius")

        distances = distance_matrix(X, self.samples_train)**2
        alpha_ = np.exp(- self.gamma * distances**2.)
        del distances
        if self.class_weight == 'balanced':
            sample_weights = compute_sample_weight('balanced', y=self.sample_classes_train)
            alpha_ *= sample_weights.T

        #
        norm_ = alpha_.sum(axis=1, keepdims=True)
        norm_[norm_==0] = 1
        alpha_ /= norm_
        #
        self.alpha = alpha_

        new_trans = alpha_ @ self.translation
        return X + new_trans

    def search_neighbors(self, X, k_nn):
        """
        Neighbors search
        :param X: array, samples
        :return: distances, mask
        """
        if np.array_equal(X, self.saved_X) and self.saved_knn == k_nn:
            if self.verbose:
                print("--- Using pre-computed neighbors distances")
            return self.distances, self.mask
        elif np.array_equal(X, self.saved_X) and self.saved_knn > k_nn:
            if self.verbose:
                print(f"--- Using pre-computed neighbors distances, --- k={k_nn}")
            return self.distances[:,:k_nn], self.mask[:,:k_nn]
        else:
            if self.verbose:
                print("--- Searching for neighbors with sklearn")
            self.distances, self.mask = self.k_nn_model.kneighbors(X, n_neighbors=k_nn)
            # self.distances_sq = self.distances ** 2
            self.saved_X = X.copy()
            self.saved_knn = k_nn
            return self.distances, self.mask

    def transform_approx(self, X, k_nn=None, gamma=None, _local_gamma=False, k_nn_local_gamma=None, loocv=False):
        """ transform X.
        :param X: array, new samples
        :return: array, new transformed samples
        """
        if loocv:
            if np.array_equal(X, None):
                X = self.samples_train
            assert np.array_equal(X, self.samples_train), "Cross Validation have to be computed on the training data"

        if k_nn == None:
            k_nn = self.k_nn
        #
        N_new = X.shape[0]
        M = self.samples_train.shape[0]

        assert X.shape[1] == self.samples_train.shape[1]
        #
        distances, mask = self.search_neighbors(X, k_nn=k_nn)
        #
        if gamma==None:
            gamma = self.gamma

        if gamma != 'optimal':
            print(f"Using gamma = {gamma}")
            if _local_gamma:
                r = distances[:,k_nn_local_gamma]
                # r = np.log(1 + distances[:,k_nn_local_gamma])
                gamma = gamma / r
                self.local_gamma = gamma.copy()
                alpha_ = np.exp(- np.float128(gamma[:,None] * distances**2.))
            else:
                alpha_ = np.exp(- np.float128(gamma * distances**2.))
            # np.exp(np.log(gamma) - gamma * distances**2.)
            if loocv:
                alpha_[:, 0] = 0
            self.alpha_unormed = alpha_.copy()
            #
            norm_ = alpha_.sum(axis=1, keepdims=True)
            norm_[norm_ == 0] = 1  # To exclude divide by 0 ToDo: Verify, this should not happen.
            alpha_ /= norm_
            # self.norm_sum = norm_

        elif gamma == 'optimal':
            d = X.shape[1]
            ind = np.zeros_like(distances)
            ind[:, :] = np.arange(1, k_nn + 1)
            # self.ind = ind.copy()
            alpha_ = (1 + d / 2 - (d / (2 * k_nn ** (2 / d))) * (ind ** (1 + 2 / d) - (ind - 1) ** (1 + 2 / d))) / k_nn
            if loocv:
                alpha_[:, 0] = 0
        else:
            raise ValueError(f"gamma parameter should de int or str 'optimal', got {gamma} instead")
        #
        mask_ravel = mask.ravel()
        # Deal with unbalanced classes:
        if self.class_weight =='balanced':
            self.sample_weights = compute_sample_weight('balanced', y=self.sample_classes_train)
            alpha_ones = self.sample_weights[mask_ravel].reshape(N_new, k_nn)
            alpha_ones /= alpha_ones.sum(axis=1, keepdims=True)
            alpha_ones *= k_nn
            alpha_ *= alpha_ones
            alpha_ /= alpha_.sum(axis=1, keepdims=True)
        else:
            self.sample_weights = np.ones([M]) # TODO: verify and add to exact method
        #
        self.weight_matrix = alpha_.copy()
        #
        # Re-arrange with sparse matrix for fast computation
        row = np.arange(0, N_new, 1)
        row = np.repeat(row, k_nn, axis=0)
        col = mask_ravel
        sparse_mat = coo_array((alpha_.ravel(), (row, col)), shape=(N_new, M))
        #
        self.sparse_mat = sparse_mat

        return sparse_mat @ self.sample_targets

    def transform_approx_ind(self, X, k_nn=None, loocv=False):
        """ transform X.
        :param X: array, new samples
        :return: array, new transformed samples
        """
        if loocv:
            if np.array_equal(X, None):
                X = self.samples_train
            assert np.array_equal(X, self.samples_train)

        if k_nn == None:
            k_nn = self.k_nn
        #
        N_new = X.shape[0]
        M = self.samples_train.shape[0]
        d = X.shape[1]

        assert X.shape[1] == self.samples_train.shape[1]
        #
        distances, mask = self.search_neighbors(X, k_nn=k_nn)
        ind = np.zeros_like(distances)
        ind[:,:] = np.arange(1, k_nn+1)
        self.ind = ind.copy()
        #
        alpha_ = (1 + d / 2 - (d / (2 * k_nn ** (2 / d))) * (ind ** (1 + 2 / d) - (ind - 1) ** (1 + 2 / d))) / k_nn
        #
        if loocv:
            alpha_[:,0] = 0

        self.alpha_unormed = alpha_.copy()
        #
        mask_ravel = mask.ravel()
        # Deal with unbalanced classes:
        if self.class_weight =='balanced':
            self.sample_weights = compute_sample_weight('balanced', y=self.sample_classes_train)
            alpha_ones = self.sample_weights[mask_ravel].reshape(N_new, k_nn)
            alpha_ones /= alpha_ones.sum(axis=1, keepdims=True)
            alpha_ones *= k_nn
            alpha_ *= alpha_ones
            alpha_ /= alpha_.sum(axis=1, keepdims=True)
        else:
            self.sample_weights = np.ones([M]) # TODO: verify and add to exact method

        # Re-arrange with sparse matrix for fast computation
        row = np.arange(0, N_new, 1)
        row = np.repeat(row, k_nn, axis=0)
        col = mask_ravel
        sparse_mat = coo_array((alpha_.ravel(), (row, col)), shape=(N_new, M))
        #
        # translations = sparse_mat @ self.translation
        self.sparse_mat = sparse_mat

        return sparse_mat @ self.sample_targets

    def transform_approx_local_on_y(self, X, k_nn=None, gamma=None, loocv=False, on_targets=False):
        """ transform X.
        :param X: array, new samples
        :return: array, new transformed samples
        """
        if loocv:
            if np.array_equal(X, None):
                X = self.samples_train
            assert np.array_equal(X, self.samples_train)
        if k_nn == None:
            k_nn = self.k_nn
        #
        N_new = X.shape[0]
        M = self.samples_train.shape[0]

        assert X.shape[1] == self.samples_train.shape[1]
        #
        distances, mask = self.search_neighbors(X, k_nn=k_nn)
        #
        if gamma==None:
            gamma = self.gamma
        # print(f"Using gamma = {gamma}")
        print(f"Using scaling_factor = {1/gamma}")

        self.local_gamma_y = gamma / self.r_y
        gamma /= self.r_y[mask]
        self.local_gamma = gamma.copy()
        alpha_ = np.exp(- np.float128(gamma * distances**2.))
        #
        if loocv:
            alpha_[:,0] = 0
        self.alpha_unormed = alpha_.copy()
        #
        norm_ = alpha_.sum(axis=1, keepdims=True)
        norm_[norm_ == 0] = 1 # To exclude divide by 0 ToDo: Verify, this should not happen.
        alpha_ /= norm_
        # self.norm_sum = norm_

        # alpha_[alpha_ < self.threshold] = 0
        # self.alpha = alpha_
        #
        mask_ravel = mask.ravel()
        # Deal with unbalanced classes:
        if self.class_weight == 'balanced':
            self.sample_weights = compute_sample_weight('balanced', y=self.sample_classes_train)
            alpha_ones = self.sample_weights[mask_ravel].reshape(N_new, k_nn)
            alpha_ones /= alpha_ones.sum(axis=1, keepdims=True)
            alpha_ones *= k_nn
            alpha_ *= alpha_ones
        else:
            self.sample_weights = np.ones([M]) # TODO: verify and add to exact method

        # Re-arrange with sparse matrix for fast computation
        row = np.arange(0, N_new, 1)
        row = np.repeat(row, k_nn, axis=0)
        col = mask_ravel
        sparse_mat = coo_array((alpha_.ravel(), (row, col)), shape=(N_new, M))
        #
        # translations = sparse_mat @ self.translation
        self.sparse_mat = sparse_mat

        # return X + translations
        if on_targets:
            return sparse_mat @ self.sample_targets
        else:
            return X + sparse_mat @ self.translation


    def transform_approx_local_on_both_x_and_y(self, X, k_nn=None, gamma=None, k_nn_local_gamma=10, loocv=False, on_targets=False):
        """ transform X.
        :param X: array, new samples
        :return: array, new transformed samples
        """
        if k_nn_local_gamma==10:
            warnings.warn("k_nn_local_gamma is set to 10 (default value)")
        if loocv:
            if np.array_equal(X, None):
                X = self.samples_train
            assert np.array_equal(X, self.samples_train)

        if k_nn == None:
            k_nn = self.k_nn
        #
        N_new = X.shape[0]
        M = self.samples_train.shape[0]

        assert X.shape[1] == self.samples_train.shape[1]
        #
        distances, mask = self.search_neighbors(X, k_nn=k_nn)
        #
        if gamma==None:
            gamma = self.gamma
        # print(f"Using gamma = {gamma}")
        print(f"Using scaling_factor = {1/gamma}")

        r_x = distances[:, k_nn_local_gamma]
        self.local_gamma_y = gamma / (self.r_y)
        gamma /= np.sqrt(self.r_y[mask] * r_x[:,None])
        self.local_gamma = gamma.copy()
        alpha_ = np.exp(- np.float128(gamma * distances**2.))
        #
        if loocv:
            alpha_[:,0] = 0
        #
        self.alpha_unormed = alpha_.copy()
        #
        norm_ = alpha_.sum(axis=1, keepdims=True)
        norm_[norm_ == 0] = 1 # To exclude divide by 0 ToDo: Verify, this should not happen.
        alpha_ /= norm_
        # self.norm_sum = norm_

        # alpha_[alpha_ < self.threshold] = 0
        # self.alpha = alpha_
        #
        mask_ravel = mask.ravel()
        # Deal with unbalanced classes:
        if self.class_weight == 'balanced':
            self.sample_weights = compute_sample_weight('balanced', y=self.sample_classes_train)
            alpha_ones = self.sample_weights[mask_ravel].reshape(N_new, k_nn)
            alpha_ones /= alpha_ones.sum(axis=1, keepdims=True)
            alpha_ones *= k_nn
            alpha_ *= alpha_ones
        else:
            self.sample_weights = np.ones([M]) # TODO: verify and add to exact method

        # Re-arrange with sparse matrix for fast computation
        row = np.arange(0, N_new, 1)
        row = np.repeat(row, k_nn, axis=0)
        col = mask_ravel
        sparse_mat = coo_array((alpha_.ravel(), (row, col)), shape=(N_new, M))
        #
        # translations = sparse_mat @ self.translation
        self.sparse_mat = sparse_mat

        # return X + translations
        if on_targets:
            return sparse_mat @ self.sample_targets
        else:
            return X + sparse_mat @ self.translation

    def transform_approx_umap_like(self, X, k_nn=None, gamma=None, k_nn_local_gamma=10, loocv=False, on_targets=False):
        """ transform X.
        :param X: array, new samples
        :return: array, new transformed samples
        """
        if k_nn_local_gamma==10:
            warnings.warn("k_nn_local_gamma is set to 10 (default value)")
        if loocv:
            if np.array_equal(X, None):
                X = self.samples_train
            assert np.array_equal(X, self.samples_train)

        if k_nn == None:
            k_nn = self.k_nn
        #
        N_new = X.shape[0]
        M = self.samples_train.shape[0]

        assert X.shape[1] == self.samples_train.shape[1]
        #
        distances, mask = self.search_neighbors(X, k_nn=k_nn)
        #
        if gamma==None:
            gamma = self.gamma
        # print(f"Using gamma = {gamma}")
        print(f"Using scaling_factor = {1/gamma}")

        r_x = distances[:, k_nn_local_gamma]
        local_gamma_y = gamma / self.r_y[mask]
        local_gamma_x = gamma / r_x[:,None]

        self.local_gamma = gamma.copy()
        alpha_y = np.exp(- np.float128(local_gamma_y * distances**2.))
        alpha_x = np.exp(- np.float128(local_gamma_x * distances**2.))
        alpha_ = alpha_x + alpha_y - alpha_x * alpha_y
        #
        if loocv:
            alpha_[:,0] = 0
        #
        self.alpha_unormed = alpha_.copy()
        #
        norm_ = alpha_.sum(axis=1, keepdims=True)
        norm_[norm_ == 0] = 1 # To exclude divide by 0 ToDo: Verify, this should not happen.
        alpha_ /= norm_
        # self.norm_sum = norm_

        # alpha_[alpha_ < self.threshold] = 0
        # self.alpha = alpha_
        #
        mask_ravel = mask.ravel()
        # Deal with unbalanced classes:
        if self.class_weight == 'balanced':
            self.sample_weights = compute_sample_weight('balanced', y=self.sample_classes_train)
            alpha_ones = self.sample_weights[mask_ravel].reshape(N_new, k_nn)
            alpha_ones /= alpha_ones.sum(axis=1, keepdims=True)
            alpha_ones *= k_nn
            alpha_ *= alpha_ones
        else:
            self.sample_weights = np.ones([M]) # TODO: verify and add to exact method

        # Re-arrange with sparse matrix for fast computation
        row = np.arange(0, N_new, 1)
        row = np.repeat(row, k_nn, axis=0)
        col = mask_ravel
        sparse_mat = coo_array((alpha_.ravel(), (row, col)), shape=(N_new, M))
        #
        # translations = sparse_mat @ self.translation
        self.sparse_mat = sparse_mat

        if on_targets:
            return sparse_mat @ self.sample_targets
        else:
            return X + sparse_mat @ self.translation


    def transform_on_target(self, X, k_nn=None, gamma=None, _local_gamma=False, k_nn_local_gamma=None, loocv=False):
        """ transform X.
        :param X: array, new samples
        :return: array, new transformed samples
        """
        if k_nn == None:
            k_nn = self.k_nn
        #
        N_new = X.shape[0]
        M = self.samples_train.shape[0]

        assert X.shape[1] == self.samples_train.shape[1]
        #
        distances, mask = self.search_neighbors(X, k_nn=k_nn)
        #
        if gamma==None:
            gamma = self.gamma
        # print(f"Using gamma = {gamma}")
        print(f"Using scaling_factor = {1/gamma}")

        if _local_gamma:
            r = distances[:,k_nn_local_gamma]
            gamma = gamma / r
            self.local_gamma = gamma.copy()
            alpha_ = np.exp(- np.float128(gamma[:,None] * distances**2.))
        else:
            alpha_ = np.exp(- np.float128(gamma * distances**2.))
        #
        if loocv:
            assert np.array_equal(X, self.samples_train)
            alpha_[:, 0] = 0
        self.alpha_unormed = alpha_.copy()
        #
        norm_ = alpha_.sum(axis=1, keepdims=True)
        norm_[norm_ == 0] = 1 # To exclude divide by 0 ToDo: Verify, this should not happen.
        alpha_ /= norm_
        # self.norm_sum = norm_

        # alpha_[alpha_ < self.threshold] = 0
        # self.alpha = alpha_
        #
        mask_ravel = mask.ravel()
        # Deal with unbalanced classes:
        if self.class_weight == 'balanced':
            self.sample_weights = compute_sample_weight('balanced', y=self.sample_classes_train)
            alpha_ones = self.sample_weights[mask_ravel].reshape(N_new, k_nn)
            alpha_ones /= alpha_ones.sum(axis=1, keepdims=True)
            alpha_ones *= k_nn
            alpha_ *= alpha_ones
        else:
            self.sample_weights = np.ones([M]) # TODO: verify and add to exact method

        # Re-arrange with sparse matrix for fast computation
        row = np.arange(0, N_new, 1)
        row = np.repeat(row, k_nn, axis=0)
        col = mask_ravel
        sparse_mat = coo_array((alpha_.ravel(), (row, col)), shape=(N_new, M))
        #
        targets = sparse_mat @ self.sample_targets

        self.sparse_mat = sparse_mat

        return targets


    def fit_cross_val(self, X, y, gamma_to_test, reduce='max'):
        """
        Cross validation score for dataset
        :param X: samples
        :param y: classes
        :param gamma_to_test: tuple, list or array of gamma to test
        :param reduce: reduction to apply along the n-dimensions (n=X.shape[1])
        :return: cross validation score (reduced or not)
        """
        # self.fit(X, y)
        error = []
        for gamma in gamma_to_test:
            X_tr = self.transform(X, gamma=gamma)
            error.append(self.compute_cross_val_err(X_tr, y, reduce=reduce))

        return error

    def compute_cross_val_err(self, X, y, reduce='max'):
        down = (1 - 1 / np.sum(self.alpha_unormed, axis=1)) ** 2
        j = (self.sample_weights[:, None] * np.square(self.ordered_targets[y] - X) # TODO: check for sample_w
                   /
                   down[:, None])
        j[j == np.inf] = 0
        J = np.nansum(j, axis=0)
        if reduce==None:
            return J
        elif reduce=='max':
            return np.max(J)
        elif reduce=='mean':
            return np.mean(J)
        else:
            raise ValueError("Invalid param reduce. Valid values are None, max and mean")

    def loo_transform(self, X=None, gamma=None):
        """
        leave-one-out transformation for.
        Transform m samples using the m-1 other samples.
        :param X: optional, learning samples to transform. Must be same as X given in fit method.
        :return: transformed samples
        """
        if np.array_equal(X, None):
            X = self.samples_train
        assert np.array_equal(X, self.samples_train)

        k_nn = self.k_nn
        #
        N_new = X.shape[0]
        M = self.samples_train.shape[0]

        if gamma==None:
            gamma = self.gamma
        print(f"Using scaling_factor = {1/gamma}")
        distances, mask = self.search_neighbors(X, k_nn=k_nn)
        #
        if gamma == None:
            gamma = self.gamma
        # print(f"Using gamma = {gamma}")
        print(f"Using scaling_factor = {1 / gamma}")
        alpha_ = np.exp(- np.float128(gamma * distances ** 2.))
        # np.exp(np.log(gamma) - gamma * distances**2.)
        #
        self.alpha_unormed = alpha_.copy()
        # alpha_[np.diag_indices_from(alpha_[:k_nn])] = 0
        # This is the only difference, put weight to itself to be 0, i.e. simulate we don't know its target
        # TODO: implement it with main transform method as an option
        alpha_[:,0] = 0
        self.alpha_tmp = alpha_.copy()
        #
        norm_ = alpha_.sum(axis=1, keepdims=True)
        norm_[norm_ == 0] = 1 # To exclude divide by 0 ToDo: Verify, this should not happen.
        alpha_ /= norm_
        # #
        mask_ravel = self.mask.ravel()
        # Deal with unbalanced classes:
        if self.class_weight == 'balanced':
            self.sample_weights = compute_sample_weight('balanced', y=self.sample_classes_train)
            alpha_ones = self.sample_weights[mask_ravel].reshape(N_new, k_nn)
            alpha_ones /= alpha_ones.sum(axis=1, keepdims=True)
            alpha_ones *= k_nn
            alpha_ *= alpha_ones
        else:
            self.sample_weights = np.ones([M]) # TODO: verify and add to exact method

        # Re-arrange with sparse matrix for fast computation
        row = np.arange(0, N_new, 1)
        row = np.repeat(row, k_nn, axis=0)
        col = mask_ravel
        sparse_mat = coo_array((alpha_.ravel(), (row, col)), shape=(N_new, M))
        # #
        # sparse_mat = self.sparse_mat
        translations = sparse_mat @ self.translation

        return X + translations

    def fit_transform(self, X, y, on_targets=True):
        """
        Fit and transform X according to y.
        :param X: array, shape (n_samples, n_features). Samples to fit
        :param y: array, shape (n_samples). Target labels
        :return:
        """
        return self.fit(X, y).transform(X, on_targets=on_targets)

    # def to_json(self, fp):
    #     model_dict = serialize_model(self)
    #     if '.json' in fp:
    #         raise NameError("Do not put .json into fp")
    #     fp += 'NonParametricKernel_only.json'
    #     with open(fp, 'w') as model_json:
    #         json.dump(model_dict, model_json)
    #     #

    # def from_json(self, fp):
    #     # Todo: See how to load the instance
    #     with open(fp, 'r') as model_json:
    #         model_dict = json.load(model_json)
    #         self.NonParametricKernel = deserialize_model(model_dict['npk'])

    def extract_k_nearest_neighbors(self,k):
        """
        Extract the k-nearest learning samples neighbors in the similarity sense.
        :param k: int, number of learning samples to extract
        :return:
        """
        return self.mask[:,:k], self.weight_matrix[:,:k]



















# Copyright UCA/CNRS/Inria
# Contributor(s): Cedric Dubois 2023
#
# cedric.dubois@inria.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import time
import tqdm
from sklearn.metrics import balanced_accuracy_score, accuracy_score
import matplotlib.pyplot as plt
import numpy as np

def search_k_opt(model, X, y, k_min=1, k_max=200, step_k=10,
                 cv_loo=True, plot_bacc=True,
                 save_fig=False, save_file_name="./",
                 show_plot=False, knn=False):
    t0 = time.time()
    if knn:
        y_pred = model.predict_knn(X, k=k_max, use_precomputed_knn=False, cv_loo=cv_loo, ).cpu()
    else:
        y_pred = model.predict(X, k=k_max, use_precomputed_knn=False, cv_loo=cv_loo, ).cpu()
    print(f"--- CV fitting time for k_max={k_max}: {np.round(time.time() - t0, 1)} s ---")
    print(f"ACC : {100 * accuracy_score(y.cpu(), y_pred)}")
    print(f"B-ACC : {100 * balanced_accuracy_score(y.cpu(), y_pred)}")
    #
    print(f"--- Computing prediction for all k from {k_max} to {k_min} with step {step_k} \n")
    k_range = np.arange(k_max, k_min, -step_k)
    acc = []
    bacc = []
    n_steps = k_range.shape[0]
    y_pred_all_k = np.zeros([y.shape[0], n_steps])
    i = 0
    for k in tqdm.tqdm(k_range):
        if knn:
            y_pred = model.predict_knn(X, k=k, use_precomputed_knn=True, cv_loo=cv_loo).cpu().numpy()
        else:
            y_pred = model.predict(X, k=k, use_precomputed_knn=True, cv_loo=cv_loo).cpu().numpy()
        y_pred_all_k[:, i] = y_pred
        acc.append(accuracy_score(y.cpu(), y_pred))
        bacc.append(balanced_accuracy_score(y.cpu(), y_pred))
        i += 1
    k_opt = k_range[np.argmax(bacc)].item()
    i_opt = np.argmax(bacc)
    print(f"Optimal k (based on B-ACC): {k_opt}")
    print(f"B-ACC max = {100 * bacc[i_opt]}")
    print(f"ACC max = {100 * acc[i_opt]}")
    np.save(save_file_name + f"y_pred_from_{k_min}_to_{k_max}_each_{step_k}_steps_k_opt_{k_opt}.npy", y_pred_all_k)
    #
    if show_plot or save_fig:
        plt.figure(dpi=300)
        plt.plot(k_range, acc, c='royalblue', ls='--', label='$\\mathrm{CV_{loo}}$ ACC')
        if plot_bacc:
            plt.plot(k_range, bacc, c='firebrick', label='$\\mathrm{CV_{loo}}$ B-ACC')
        plt.xlabel('$k$')
        plt.legend()
        plt.grid()
        if save_fig:
            plt.savefig(save_file_name + f'from_{k_min}_to_{k_max}_each_{step_k}_steps.pgf', dpi=300)
            plt.savefig(save_file_name + f'from_{k_min}_to_{k_max}_each_{step_k}_steps.png', dpi=300)
            plt.savefig(save_file_name + f'from_{k_min}_to_{k_max}_each_{step_k}_steps.pdf', dpi=300)
        if show_plot:
            plt.show()
        plt.close()
    return model

class TorchStandardScaler:
    #
    def fit(self, x):
        self.mean = x.mean(0, keepdim=True)
        self.std = x.std(0, unbiased=False, keepdim=True)
        return self

    def transform(self, x):
        x -= self.mean
        x /= (self.std + 1e-7)
        return x

    def fit_transform(self, x):
        return self.fit(x).transform(x)


class TorchMinMaxScaler:
    def __init__(self, min=0, max=1):
        self.min = min
        self.max = max

    def fit(self, x):
        self.x_min = x.min(axis=0, keepdim=True).values
        self.x_max = x.max(axis=0, keepdim=True).values
        return self

    def transform(self, x):
        X_std = (x - self.x_min) / (self.x_max - self.x_min)
        return X_std * (self.max - self.min) + self.min

    def fit_transform(self, x):
        return self.fit(x).transform(x)



# Copyright UCA/CNRS/Inria
# Contributor(s): Cedric Dubois 2023
#
# cedric.dubois@inria.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from pykeops.torch import LazyTensor
import torch

# Detect whether to use CUDA or CPU
USE_CUDA = torch.cuda.is_available()
DEVICE = 'cuda' if USE_CUDA else 'cpu'
tensor = torch.cuda.FloatTensor if USE_CUDA else torch.FloatTensor
# if USE_CUDA:
#     torch.set_default_tensor_type('torch.cuda.FloatTensor')

class WKNNClassifier():
    """
    PyKeops based W-k-NN classifier

    Parameters:
    -----------
    weights : str
        Weights used for the weighted k-NN algorithm. Default is 'optimal'.
    k : int
        Number of neighbors to consider. Default is 10.
    targets_def : str
        Definition of the targets to use. Default is 'optimal' => w-k-nn.
        Can be 'provided' for user defined targets, note the classifier is no more a w-k-nn in this case.
    class_weight : str or None
        Method used to compute class weights for w-k-NN. Default is None, can be 'balanced' to compensate (inversely proportional)
    gamma : float
        If use the wNN. Default is 1.
    verbose : int
        Verbosity level of the classifier. Default is 0.
    matrix_device : str
        Device where the symbolic matrix is located. Default is 'cuda'.

    Methods:
    --------
    fit(X, y, targets=None)
        Fit the classifier on the input data.

    predict_knn(X, k=None, use_precomputed_knn=False, cv_loo=False)
        Predict the labels of the input samples using the w-k-NN method.

    search_neighbors(X, k, use_precomputed_knn=False, cv_loo=False)
        Search the k Nearest-Neighbors of the input samples.

    save_model(fp)
        Save the Nearest-Neighbors to disk.

    load_model(fp, k_max)
        Load the Nearest-Neighbors from disk.

    Attributes:
    -----------
    k : int
        number of Nearest-Neighbors to consider.
    weights : str
        type of weights for the weighted-k-NN algorithm (w-k-NN or w-NN).
    gamma : float
        scaling factor for w-NN.
    targets_def : str
        type of target values.
    verbose : int
        level of verbosity.
    class_weight : str
        type of weighting for the classes.
    device : str
        device on which the distance matrix will be computed.
    saved_knn : int
        number of Nearest-Neighbors kept in memory.
    """
    def __init__(self,
                 weights='optimal',
                 k=10,
                 targets_def='optimal',
                 class_weight=None,
                 gamma=1,
                 verbose=0,
                 matrix_device='cuda',
                 ):
        self.k = k
        self.weights = weights
        self.gamma = gamma
        self.targets_def = targets_def
        self.verbose = verbose
        self.class_weight = class_weight
        self.device = matrix_device

        # torch.cuda.set_device(torch.device(DEVICE, 0))

        self.saved_knn = 0

    def search_neighbors(self, X, k, use_precomputed_knn=False, cv_loo=False):
        """
        Find the k nearest neighbors of a set of samples.

        Parameters:
        -----------
        X : array-like of shape (n_samples, d_features)
            Samples to find the k nearest neighbors of.
        k : int
            The number of neighbors to consider.
        use_precomputed_knn : bool, optional (default=False)
            Whether to use precomputed nearest neighbors distances (kept in memory).
        cv_loo : bool, optional (default=False)
            Whether to use leave-one-out cross-validation (LOO-CV).

        Returns:
        --------
        ind_knn : tensor of shape (n_samples, k)
            Indices of the k nearest neighbors of each point in X.
        """
        if k<1:
            raise ValueError(f"k must be > 0, got {k} instead")
        # If saved_knn=0 (i.e. first pass), then compute the NN search.
        if use_precomputed_knn and self.saved_knn == k:
            if self.verbose:
                print("--- Using pre-computed neighbors distances")
            if cv_loo:
                return self.ind_knn[:,1:]
            else:
                return self.ind_knn
        elif use_precomputed_knn and self.saved_knn > k:
            if self.verbose:
                print(f"--- Using pre-computed neighbors distances, with k={k} ---")
            if cv_loo:
                return self.ind_knn[:,1:k+1]
            else:
                return self.ind_knn[:, :k]
        else:
            if self.verbose:
                print("--- Searching for neighbors with symbolic matrix ---")
            self.saved_knn = k
            D_ij = ((X - self.X_j) ** 2).sum(-1)  # (M, N) symbolic matrix of squared L2 distances
            self.D = D_ij
            if cv_loo:
                self.ind_knn = D_ij.argKmin(k + 1, dim=1)  # Samples <-> Dataset, (M, K)
                return self.ind_knn[:,1:]
            else:
                self.ind_knn = D_ij.argKmin(k, dim=1)  # Samples <-> Dataset, (M, K)
                return self.ind_knn

    def save_model(self, fp):
        """
        Save nearest neighbours
        :param fp:
        :return:
        """
        torch.save(self.ind_knn, fp)

    def load_model(self, fp, k_max):
        """
        Load nearest neighbours
        :param fp:
        :return:
        """
        self.ind_knn = torch.load(fp)
        self.saved_knn = k_max

    def fit(self, X, y, targets=None):
        """
        Fit the model to the training data.

        Parameters
        ----------
        X : numpy.ndarray or torch.Tensor
            The input samples, of shape (n_samples, d_features).
        y : numpy.ndarray or torch.Tensor
            The corresponding labels, of shape (n_samples,).
        targets : numpy.ndarray or torch.Tensor, optional
            The target output values for each class, of shape (n_classes, n_classes). If not provided,
            optimal targets (least square) will be used.

        Returns
        -------
        self : The fitted model instance.

        Raises
        ------
        ValueError
            If `targets_def` is not 'optimal' or 'provided'.
        """

        self.n_samples_ = X.shape[0]
        self.d = X.shape[1]
        self.n_classes_ = torch.unique(y).shape[0]

        # Set the target values based on the `targets_def` parameter
        if self.targets_def == 'optimal':
            self.targets = torch.eye(self.n_classes_, device='cuda') # optimal targets (least square)
        elif self.targets_def == 'provided':
            self.targets = targets
        else:
            raise ValueError(f"{self.targets_def} must be 'optimal' or 'provided'")

        self.labels = y.to(self.device)     # Convert the labels to CPU or GPU
        self.y_targets = self.targets[tensor.long(y)]

        # Create a train LazyTensor (1, n, d) from the input samples
        self.X_j = LazyTensor(X[None, :, :])

        # Compute class weights if `class_weight` is set to 'balanced'
        if self.class_weight == 'balanced':
            _, counts = torch.unique(y, return_counts=True)
            self.cls_weights = 1 / counts

        return self

    def predict_knn(self, X, k=None, use_precomputed_knn=False, cv_loo=False):
        """
        Predict the labels with the classical (uniform) k-NN
        :param X: samples of unknown labels
        :param k: number of neighbors to consider
        :param use_precomputed_knn: Whether to use precomputed k-NN distances
        :param cv_loo: Whether to use leave-one-out cross-validation
        :return: predicted labels
        """
        # Set k to the maximum number of samples if not provided
        k_max = self.n_samples_
        if k == None:
            k = self.k
        if k > k_max:
            k = k_max

        # Test Lazy Tensor (m, 1, d)
        X_i = LazyTensor(X[:, None, :])

        # Find k nearest neighbors and retrieve their labels
        ind_knn = self.search_neighbors(X_i, k, use_precomputed_knn=use_precomputed_knn, cv_loo=cv_loo)
        lab_knn = self.labels[ind_knn]  # (m_test, k)

        # Compute the most likely label among the k neighbors
        y_knn, _ = lab_knn.mode()

        return y_knn

    def transform(self, X, k=None, gamma=None, use_precomputed_knn=False, cv_loo=False):
        """
       Transform samples (NN-kernel)

       :param X: input data (shape: [m_samples, d_features])
       :param k: number of nearest neighbors (default: self.k)
       :param gamma: scale parameter for the Gaussian kernel (default: self.gamma)
       :param use_precomputed_knn: Whether to use precomputed k-NN distances
       :param cv_loo: Whether to use leave-one-out cross-validation

       :return: transformed data (shape: [m_samples, n_classes])
       """
        k_max = self.n_samples_

        if k == None:
            k = self.k
        if k > k_max:
            k = k_max

        if gamma == None:
            gamma = self.gamma

        d = self.d
        assert d == X.shape[1], "Number of features in input data does not match number of features in training data"

        X_i = LazyTensor(X[:, None, :])   # (M, 1, d) test set
        m_samples = X.shape[0]

        if self.weights=='optimal':
            # compute k nearest neighbors indices and weights
            ind_knn = self.search_neighbors(X_i, k, use_precomputed_knn=use_precomputed_knn, cv_loo=cv_loo)
            ind = torch.arange(1, k + 1, device=self.device) # large memory usage for large n_samples
            weights = (1 + d / 2 - (d / (2 * k ** (2 / d))) * (ind ** (1 + 2 / d) - (ind - 1) ** (1 + 2 / d))) / k

            # transform data
            x_tr = tensor(torch.Size((m_samples, self.n_classes_)))
            for i in range(self.n_classes_):
                x_tr[:, i] = ((self.labels[ind_knn] == i) * weights[None, :]).sum(axis=-1)

        elif self.weights=='Gaussian' or self.weights=='gaussian':
            if cv_loo:
                raise NotImplementedError("No yet implemented")
            else:
                D_ij = ((X_i - self.X_j) ** 2).sum(-1)  # (M, N) symbolic matrix of squared L2 distances
                self.D = D_ij
                K_ij = (- gamma * D_ij).exp()
                x_tr = (K_ij @ self.targets[tensor.long(self.labels)]) / K_ij.sum(1)
        else:
            raise NotImplementedError("No yet implemented")

        if USE_CUDA:
            torch.cuda.synchronize()

        if self.class_weight == 'balanced':
            # Re-balance the weights and normalize
            x_tr *= self.cls_weights
            x_tr /= x_tr.sum(axis=1, keepdims=True)

        return x_tr

    def predict(self,X, **kwargs):
        """
        Predict labels with w-k-nn
        :param X: input samples of unknown labels
        :param kwargs: arguments passed to the `transform` method:
           :param k: number of nearest neighbors (default: self.k)
           :param gamma: scale parameter for the Gaussian kernel (default: self.gamma)
           :param use_precomputed_knn: Whether to use precomputed k-NN distances
           :param cv_loo: Whether to use leave-one-out cross-validation

        :return: predicted labels
        """
        return self.transform(X, **kwargs).argmax(1)
